import glob
import os
from collections import defaultdict

from nbgrader.api import Gradebook, MissingEntry
from nbgrader.exchange.default.collect import ExchangeCollect, groupby
from nbgrader.utils import check_mode
from traitlets import Bool

from .exchange import E2xExchange


class E2xExchangeCollect(E2xExchange, ExchangeCollect):
    update = Bool(
        False, help="Update existing submissions with ones that have newer timestamps."
    ).tag(config=True)

    before_duedate = Bool(
        False,
        help=(
            "Collect the last submission before due date or the last submission\n"
            "if no submission before due date."
        ),
    ).tag(config=True)

    check_owner = Bool(
        default_value=True,
        help=(
            "Whether to cross-check the student_id with the UNIX-owner of the submitted directory."
        ),
    ).tag(config=True)

    def init_submissions(self):
        if self.personalized_inbound:
            self.log.info("Collecting from restricted submit dirs")
            submit_dirs = [
                username
                for username in os.listdir(self.inbound_path)
                if "+" not in username
                and os.path.isdir(os.path.join(self.inbound_path, username))
            ]
            self.log.info(f"Submission dirs: {submit_dirs}")

            usergroups = defaultdict(list)
            records = []

            for user in submit_dirs:
                submit_path = os.path.join(self.inbound_path, user)
                self.log.info(f"Assignment id: {self.coursedir.assignment_id}")
                pattern = os.path.join(
                    submit_path, f"{user}+{self.coursedir.assignment_id}+*"
                )
                user_records = [self._path_to_record(f) for f in glob.glob(pattern)]
                self.log.info(f"{user} has {len(user_records)} submissions.")
                for i, record in enumerate(user_records):
                    user_records[i]["filename"] = os.path.join(user, record["filename"])

                usergroups.update(groupby(user_records, lambda item: item["username"]))
                records.append(user_records)

        else:
            student_id = self.coursedir.student_id if self.coursedir.student_id else "*"
            pattern = os.path.join(
                self.inbound_path,
                "{}+{}+*".format(student_id, self.coursedir.assignment_id),
            )
            records = [self._path_to_record(f) for f in glob.glob(pattern)]
            usergroups = groupby(records, lambda item: item["username"])

        return records, usergroups

    def init_src(self):
        if self.coursedir.course_id == "":
            self.fail("No course id specified. Re-run with --course flag.")

        self.course_path = os.path.join(self.root, self.coursedir.course_id)
        self.inbound_path = os.path.join(self.course_path, self.inbound_directory)
        if not os.path.isdir(self.inbound_path):
            self.fail("Course not found: {}".format(self.inbound_path))
        if not check_mode(self.inbound_path, read=True, execute=True):
            self.fail(
                "You don't have read permissions for the directory: {}".format(
                    self.inbound_path
                )
            )

        records, usergroups = self.init_submissions()

        with Gradebook(self.coursedir.db_url, self.coursedir.course_id) as gb:
            try:
                assignment = gb.find_assignment(self.coursedir.assignment_id)
                self.duedate = assignment.duedate
            except MissingEntry:
                self.duedate = None
        if self.duedate is None or not self.before_duedate:
            self.src_records = [
                self._sort_by_timestamp(v)[0] for v in usergroups.values()
            ]
        else:
            self.log.info(
                "Duedate is enabled. Collecting assignments submitted before duedate..."
            )
            self.src_records = []
            for v in usergroups.values():
                records = self._sort_by_timestamp(v)
                records_before_duedate = [
                    record for record in records if record["timestamp"] <= self.duedate
                ]
                if records_before_duedate:
                    self.src_records.append(records_before_duedate[0])
                else:
                    self.src_records.append(records[0])



import unittest

from nbformat.notebooknode import NotebookNode
from nbformat.v4 import new_code_cell, new_output

from e2xgrader.graders import CodeGrader


class TestCodeGrader(unittest.TestCase):
    def create_autograded_code_cell(
        self,
        grade_id: str,
        points: int = 5,
    ) -> NotebookNode:
        cell = new_code_cell()

        cell.metadata = {
            "nbgrader": {
                "grade": True,
                "grade_id": grade_id,
                "locked": False,
                "points": points,
                "schema_version": 3,
                "solution": False,
                "task": False,
            }
        }
        return cell

    def test_errored_cell(self):
        cell = self.create_autograded_code_cell("mycell", 10)

        cell.outputs.append(
            new_output(
                output_type="error", ename="MyError", evalue="Something went wrong."
            )
        )
        grader = CodeGrader()

        points, max_points = grader.determine_grade(cell)
        assert points == 0
        assert max_points == 10

    def test_stderrored_cell(self):
        grader = CodeGrader()
        cell = self.create_autograded_code_cell("mycell", 10)
        cell.outputs.append(new_output(output_type="stream", name="stderr"))

        points, max_points = grader.determine_grade(cell)
        assert points == 0
        assert max_points == 10

    def test_correctly_formatted_execute_result(self):
        grader = CodeGrader()
        cell = self.create_autograded_code_cell("mycell", 10)
        cell.outputs.append(
            new_output(output_type="execute_result", data={"text/plain": "5.0"})
        )
        points, max_points = grader.determine_grade(cell)
        assert points == 5
        assert max_points == 10

    def test_incorrectly_formatted_execute_result(self):
        grader = CodeGrader()
        cell = self.create_autograded_code_cell("mycell", 10)
        cell.outputs.append(
            new_output(output_type="execute_result", data={"text/plain": "five"})
        )
        points, max_points = grader.determine_grade(cell)
        assert points == max_points
        assert max_points == 10

    def test_correctly_formatted_partial_credit(self):
        grader = CodeGrader()
        cell = self.create_autograded_code_cell("mycell", 10)
        cell.outputs.append(
            new_output(
                output_type="stream",
                name="stdout",
                text="### BEGIN GRADE\n5.0\n### END GRADE\n",
            )
        )
        points, max_points = grader.determine_grade(cell)
        assert points == 5
        assert max_points == 10

    def test_incorrectly_formatted_partial_credit(self):
        grader = CodeGrader()
        cell = self.create_autograded_code_cell("mycell", 10)
        cell.outputs.append(
            new_output(
                output_type="stream",
                name="stdout",
                text="### BEGIN GRADE\nfive\n### END GRADE\n",
            )
        )
        points, max_points = grader.determine_grade(cell)
        assert points is max_points
        assert max_points == 10



def is_extra_cell(cell):
    """Returns True if the cell is a form cell."""
    if "nbgrader" not in cell.metadata:
        return False
    return "extended_cell" in cell.metadata


def is_attachment_cell(cell):
    return is_extra_cell(cell) and cell.metadata.extended_cell.type == "attachments"


def is_singlechoice(cell):
    return is_extra_cell(cell) and cell.metadata.extended_cell.type == "singlechoice"


def is_multiplechoice(cell):
    return is_extra_cell(cell) and cell.metadata.extended_cell.type == "multiplechoice"


def is_diagram(cell):
    return is_extra_cell(cell) and cell.metadata.extended_cell.type == "diagram"


def get_options(cell):
    if is_extra_cell(cell) and "options" in cell.metadata.extended_cell:
        return {
            key: value["value"]
            for key, value in cell.metadata.extended_cell["options"].items()
        }
    return dict()


def get_choices(cell):
    if is_singlechoice(cell) or is_multiplechoice(cell):
        choice = cell.metadata.extended_cell.get("choice", [])
        return [int(i) for i in choice]
    return []


def get_num_of_choices(cell):
    if is_multiplechoice(cell):
        return cell.metadata.extended_cell.num_of_choices


def get_instructor_choices(cell):
    if (
        (is_singlechoice(cell) or is_multiplechoice(cell))
        and ("source" in cell.metadata.extended_cell)
        and ("choice" in cell.metadata.extended_cell.source)
    ):
        return [int(i) for i in cell.metadata.extended_cell.source.choice]
    return []


def clear_choices(cell):
    if is_extra_cell(cell):
        cell.metadata.extended_cell.choice = []


def has_solution(cell):
    if is_singlechoice(cell) or is_multiplechoice(cell):
        return "source" in cell.metadata.extended_cell
    return False



import os
import unittest
from tempfile import TemporaryDirectory

import nbformat
from nbformat.v4 import new_notebook

from e2xgrader.server_extensions.apps.validate_assignment.validator import E2XValidator
from e2xgrader.tests.test_utils.cells import (
    new_multiplechoice_cell,
    new_singlechoice_cell,
    new_upload_cell,
)


class TestValidator(unittest.TestCase):
    def setUp(self):
        self.tmp_dir = TemporaryDirectory()
        self.validator = E2XValidator()
        # There is no need to execute the notebook
        self.validator.preprocessors = []
        self.error_msg = "You did not provide a response."

    def save_notebook(self, nb):
        nb_name = "mynb.ipynb"
        filepath = os.path.join(self.tmp_dir.name, nb_name)
        nbformat.write(nb, filepath)
        return filepath

    def test_validate_pass_singlechoice_cell(self):
        nb = new_notebook()
        nb.cells.append(new_singlechoice_cell(choice=["0"], points=4))
        filepath = self.save_notebook(nb)
        result = self.validator.validate(filepath)

        assert len(result) == 0

    def test_validate_fail_singlechoice_cell(self):
        nb = new_notebook()
        nb.cells.append(new_singlechoice_cell(points=5))
        filepath = self.save_notebook(nb)

        result = self.validator.validate(filepath)

        assert "failed" in result
        assert len(result["failed"]) == 1
        assert result["failed"][0]["error"] == self.error_msg

    def test_validate_pass_multiplechoice_cell(self):
        nb = new_notebook()
        nb.cells.append(new_multiplechoice_cell(points=5, choice=["0", "1"]))
        filepath = self.save_notebook(nb)

        result = self.validator.validate(filepath)
        assert len(result) == 0

    def test_validate_fail_multiplechoice_cell(self):
        nb = new_notebook()
        nb.cells.append(new_multiplechoice_cell(points=5))
        filepath = self.save_notebook(nb)

        result = self.validator.validate(filepath)
        assert "failed" in result
        assert len(result["failed"]) == 1
        assert result["failed"][0]["error"] == self.error_msg

    def test_validate_pass_attachment_cell(self):
        nb = new_notebook()
        nb.cells.append(
            new_upload_cell(
                points=10, attachments={"test.png": {"image/png": "byte64randomdata"}}
            )
        )
        filepath = self.save_notebook(nb)

        result = self.validator.validate(filepath)

        assert len(result) == 0

    def test_validate_fail_attachment_cell(self):
        nb = new_notebook()
        nb.cells.append(new_upload_cell(points=3))
        filepath = self.save_notebook(nb)

        result = self.validator.validate(filepath)

        assert "failed" in result
        assert len(result["failed"]) == 1
        assert result["failed"][0]["error"] == self.error_msg

    def test_pass_other_extra_cells(self):
        extra_cell = new_singlechoice_cell()
        extra_cell.metadata["extended_cell"]["type"] = "sometype"

        nb = new_notebook()
        nb.cells.append(extra_cell)
        filepath = self.save_notebook(nb)

        result = self.validator.validate(filepath)

        assert len(result) == 0

    def tearDown(self):
        self.tmp_dir.cleanup()



import sys
from argparse import ArgumentParser
from textwrap import dedent

from ..extensions import E2xExtensionManager


class Manager:
    def __init__(self):
        self.extension_manager = E2xExtensionManager()
        parser = ArgumentParser(
            description="E2X extension manager.",
            usage=dedent(
                """
                e2xgrader <command> [<args>]

                Available sub commands are:
                activate      activate a specific mode (teacher, student, student-exam)
                deactivate    deactivate all extensions"""
            ),
        )

        parser.add_argument("command", help="Subcommand to run")

        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print("Unrecognized command")
            parser.print_help()
            exit(1)
        getattr(self, args.command)()

    def activate(self):
        parser = ArgumentParser(
            description="Activate different modes",
            usage=dedent(
                """
                e2xgrader activate <mode> [--sys-prefix] [--user]

                Available modes are:
                teacher         activate the grader and all teaching extensions
                student         activate the student extensions
                student_exam    activate the student extensions in exam mode"""
            ),
        )
        # prefixing the argument with -- means it's optional
        parser.add_argument(
            "mode",
            help="Which mode to activate, can be teacher, student or student-exam",
        )
        parser.add_argument(
            "--sys-prefix",
            action="store_true",
            help="If the extensions should be installed to sys.prefix",
        )
        parser.add_argument(
            "--user",
            action="store_true",
            help="If the extensions should be installed to the user space",
        )

        args = parser.parse_args(sys.argv[2:])
        if not hasattr(self.extension_manager, f"activate_{args.mode}"):
            print("Unrecognized mode")
            parser.print_help()
            exit(1)
        sys_prefix = False
        user = False
        if args.sys_prefix:
            sys_prefix = True
        if args.user:
            user = True
        getattr(self.extension_manager, f"activate_{args.mode}")(
            sys_prefix=sys_prefix, user=user
        )

    def deactivate(self):
        parser = ArgumentParser(
            description="Deactivate extensions",
            usage=dedent("python -m e2xgrader deactivate [--sys-prefix] [--user]"),
        )
        # prefixing the argument with -- means it's optional
        parser.add_argument(
            "--sys-prefix",
            action="store_true",
            help="If the extensions should be uninstalled from sys.prefix",
        )
        parser.add_argument(
            "--user",
            action="store_true",
            help="If the extensions should be uninstalled from the user space",
        )

        args = parser.parse_args(sys.argv[2:])

        sys_prefix = False
        user = False
        if args.sys_prefix:
            sys_prefix = True
        if args.user:
            user = True
        self.extension_manager.deactivate(sys_prefix=sys_prefix, user=user)


def main():
    Manager()



from ..student.student import StudentExtension


class StudentExamExtension(StudentExtension):
    pass


def load_jupyter_server_extension(nbapp):
    """Load the e2xgrader serverextension"""
    nbapp.log.info("Loading the e2xgrader student exam serverextension")
    StudentExamExtension(parent=nbapp)



import glob
import os
import re
import shutil
from stat import (
    S_IRGRP,
    S_IROTH,
    S_IRUSR,
    S_ISGID,
    S_IWGRP,
    S_IWOTH,
    S_IWUSR,
    S_IXGRP,
    S_IXOTH,
    S_IXUSR,
)

from nbgrader.exchange.default import ExchangeReleaseFeedback
from nbgrader.utils import make_unique_key, notebook_hash

from .exchange import E2xExchange


class E2xExchangeReleaseFeedback(E2xExchange, ExchangeReleaseFeedback):
    def init_dest(self):
        """
        Create exchange feedback destination
        """

        if self.coursedir.course_id == "":
            self.fail("No course id specified. Re-run with --course flag.")

        self.course_path = os.path.join(self.root, self.coursedir.course_id)
        self.dest_path = os.path.join(self.course_path, self.feedback_directory)

        if self.personalized_feedback:
            # u+rwx, g+wx, o+wx
            self.ensure_directory(
                self.dest_path,
                (
                    S_IRUSR
                    | S_IWUSR
                    | S_IXUSR
                    | S_IWGRP
                    | S_IXGRP
                    | S_IWOTH
                    | S_IXOTH
                    | (
                        (S_IRGRP | S_IWGRP | S_ISGID)
                        if self.coursedir.groupshared
                        else 0
                    )
                ),
            )
        else:
            # 0755
            self.ensure_directory(
                self.dest_path,
                (
                    S_IRUSR
                    | S_IWUSR
                    | S_IXUSR
                    | S_IXGRP
                    | S_IXOTH
                    | (
                        (S_IRGRP | S_IWGRP | S_ISGID)
                        if self.coursedir.groupshared
                        else 0
                    )
                ),
            )

    def copy_files(self):
        """
        Overried copy_files and add personalized-feedback generation
        """
        if self.coursedir.student_id_exclude:
            exclude_students = set(self.coursedir.student_id_exclude.split(","))
        else:
            exclude_students = set()

        html_files = glob.glob(os.path.join(self.src_path, "*.html"))
        for html_file in html_files:
            if "hashcode" in html_file:
                self.log.debug("Skipping hashcode info")
                continue
            regexp = re.escape(os.path.sep).join(
                [
                    self.coursedir.format_path(
                        self.coursedir.feedback_directory,
                        "(?P<student_id>.*)",
                        self.coursedir.assignment_id,
                        escape=True,
                    ),
                    "(?P<notebook_id>.*).html",
                ]
            )

            m = re.match(regexp, html_file)
            if m is None:
                msg = "Could not match '%s' with regexp '%s'" % (html_file, regexp)
                self.log.error(msg)
                continue

            gd = m.groupdict()
            student_id = gd["student_id"]
            notebook_id = gd["notebook_id"]
            if student_id in exclude_students:
                self.log.debug("Skipping student '{}'".format(student_id))
                continue

            feedback_dir = os.path.split(html_file)[0]
            submission_dir = self.coursedir.format_path(
                self.coursedir.submitted_directory,
                student_id,
                self.coursedir.assignment_id,
            )

            if self.personalized_feedback:
                dest = os.path.join(
                    self.dest_path, student_id, self.coursedir.assignment_id
                )
                # u+rwx, g+wx, o+wx
                self.ensure_directory(
                    dest,
                    (
                        S_IRUSR
                        | S_IWUSR
                        | S_IXUSR
                        | S_IRGRP
                        | S_IXGRP
                        | S_IXOTH
                        | S_IROTH
                        | (
                            (S_IRGRP | S_IWGRP | S_ISGID)
                            if self.coursedir.groupshared
                            else 0
                        )
                    ),
                )

                dest = os.path.join(dest, notebook_id + ".html")

                self.log.info(
                    "Releasing feedback for student '{}' on assignment '{}/{}/{}' ".format(
                        student_id,
                        self.coursedir.course_id,
                        self.coursedir.assignment_id,
                        notebook_id,
                    )
                )
            else:
                timestamp = open(os.path.join(feedback_dir, "timestamp.txt")).read()
                nbfile = os.path.join(submission_dir, "{}.ipynb".format(notebook_id))
                unique_key = make_unique_key(
                    self.coursedir.course_id,
                    self.coursedir.assignment_id,
                    notebook_id,
                    student_id,
                    timestamp,
                )

                self.log.debug("Unique key is: {}".format(unique_key))
                checksum = notebook_hash(nbfile, unique_key)
                dest = os.path.join(self.dest_path, "{}.html".format(checksum))

                self.log.info(
                    "Releasing feedback for student '{}' on assignment '{}/{}/{}' ({})".format(
                        student_id,
                        self.coursedir.course_id,
                        self.coursedir.assignment_id,
                        notebook_id,
                        timestamp,
                    )
                )

            shutil.copy(html_file, dest)
            self.log.info("Feedback released to: {}".format(dest))



import typing
from collections import defaultdict
from textwrap import dedent

import pandas as pd
from e2xcore.api import E2xAPI
from e2xcore.utils.utils import get_nbgrader_config
from nbgrader.api import Grade
from traitlets import Bool, List, Unicode
from traitlets.config import LoggingConfigurable


class GradeExporter(LoggingConfigurable):
    assignments = List(
        trait=Unicode(),
        default_value=[],
        help="The assignments for which grades should be exported",
    )

    tasks = Bool(
        default_value=False,
        help="Whether to include scores per cell or not. Defaults to False.",
    )

    notebooks = Bool(
        default_value=True,
        help=dedent(
            """
            Whether to include scores per notebook or not. 
            Can only be False if tasks is False. Defaults to True.
        """
        ),
    )

    include_max_score = Bool(
        default_value=False,
        help=dedent(
            """
            Whether to include a row with the maximum score or not.
            If True a row named "max_score" will be added. Defaults to False.
        """
        ),
    )

    normalize = Bool(
        default_value=False,
        help=dedent(
            """
            Whether to divide all scores by the maximum score or not.
            Defaults to False.
        """
        ),
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.api = E2xAPI(config=get_nbgrader_config())

    def get_assignment_ids(self) -> typing.List[str]:
        if len(self.assignments) > 0:
            return self.assignments
        else:
            return [
                assignment["name"]
                for assignment in self.api.get_assignments(include_score=False)
            ]

    def get_grades(self) -> pd.DataFrame:
        data = defaultdict(lambda: defaultdict(float))
        assignments = self.get_assignment_ids()
        for assignment_id in assignments:
            for notebook in self.api.get_notebooks(assignment_id):
                notebook_id = notebook["name"]
                # Get the maximum score
                if self.include_max_score or self.normalize:
                    for key, value in self.max_score(assignment_id, notebook).items():
                        data[key].update(value)
                for submission in self.api.get_notebook_submissions(
                    assignment_id=assignment_id, notebook_id=notebook_id
                ):
                    if self.tasks:
                        grades = self.task_grades(assignment_id, submission)
                        for key, value in grades.items():
                            data[key].update(value)
                    elif self.notebooks:
                        data[(assignment_id, notebook_id)][
                            submission["student"]
                        ] = submission["score"]
                    else:
                        data[assignment_id][submission["student"]] += submission[
                            "score"
                        ]
        grades = pd.DataFrame.from_dict(data, orient="index").T
        if self.normalize:
            grades = grades.div(grades.loc["max_score"]).drop("max_score")
        return grades[sorted(grades.columns)]

    def max_score(
        self, assignment_id: str, notebook: typing.Dict[str, typing.Any]
    ) -> typing.Dict[str, typing.Dict[str, float]]:
        data = defaultdict(dict)
        assignment = self.api.get_assignment(assignment_id, include_score=False)
        if assignment["num_submissions"] < 1 or notebook["max_score"] == 0:
            return data
        if self.tasks:
            with self.api.gradebook as gb:
                for grade in gb.find_notebook(
                    name=notebook["name"], assignment=assignment_id
                ).grade_cells:
                    grade_id = self.normalize_grade_id(grade)
                    data[(assignment_id, notebook["name"], grade_id)][
                        "max_score"
                    ] = grade.max_score
        elif self.notebooks:
            data[(assignment_id, notebook["name"])]["max_score"] = notebook["max_score"]
        else:
            data[assignment_id]["max_score"] = self.api.get_assignment(
                assignment_id, include_score=False
            )["max_score"]
        return data

    def task_grades(
        self, assignment_id: str, submission: typing.Dict[str, typing.Any]
    ) -> typing.Dict[str, typing.Dict[str, float]]:
        data = defaultdict(dict)
        with self.api.gradebook as gb:
            nb = gb.find_submission_notebook_by_id(submission["id"])
            for grade in nb.grades:
                grade_id = self.normalize_grade_id(grade)
                data[(assignment_id, submission["name"], grade_id)][
                    submission["student"]
                ] = grade.score
        return data

    def normalize_grade_id(self, grade: Grade) -> str:
        return grade.name[5:] if grade.name.startswith("test_") else grade.name



from e2xcore import BaseApp
from nbgrader.apps.baseapp import NbGrader
from tornado import web
from traitlets import List, Unicode

from .apihandlers import default_handlers


class Help(NbGrader, BaseApp):
    shared_paths = List(
        trait=Unicode(),
        default_value=[],
        help="List of paths of files served via the help app",
    ).tag(config=True)

    def __init__(self, **kwargs):
        NbGrader.__init__(self, **kwargs)
        BaseApp.__init__(self, **kwargs)

    def get_static_handlers(self):
        static_handlers = [
            (
                f"/e2x/help/static/{idx}/(.*)",
                web.StaticFileHandler,
                dict(path=path, default_filename="index.html"),
            )
            for idx, path in enumerate(self.shared_paths)
        ]
        for idx, path in enumerate(self.shared_paths):
            static_handlers.append(
                (
                    f"/e2x/help/static/{idx}/(.*)",
                    web.StaticFileHandler,
                    dict(path=path, default_filename="index.html"),
                )
            )
        return static_handlers

    def load_app(self):
        self.initialize([])
        self.log.info("Loading the e2xgrader help app")

        self.update_tornado_settings(
            dict(
                e2xhelp_shared_dirs={
                    str(idx): path for idx, path in enumerate(self.shared_paths)
                }
            )
        )
        self.add_handlers(self.get_static_handlers() + default_handlers)



from nbgrader.preprocessors import NbGraderPreprocessor


class FilterCellsById(NbGraderPreprocessor):
    def filter_cells(self, cells, keyword):
        new_cells = []
        for cell in cells:
            metadata = cell["metadata"]
            if (
                ("nbgrader" in metadata)
                and ("grade_id" in metadata["nbgrader"])
                and (keyword in cell["metadata"]["nbgrader"]["grade_id"])
            ):
                new_cells.append(cell)
        return new_cells

    def preprocess(self, nb, resources):
        if "keyword" in resources:
            nb.cells = self.filter_cells(nb.cells, resources["keyword"])
        return nb, resources
